/*
 * Copyright (c) Jupyter Development Team.
 * Distributed under the terms of the Modified BSD License.
 */

module.exports = {
  testEnvironment: 'node',
  testRegex: 'lib/test/.*.spec.js[x]?$',
};
