// Copyright (c) Jupyter Development Team.
// Distributed under the terms of the Modified BSD License.

import type { Awareness } from 'y-protocols/awareness';

/**
 * The awareness interface.
 */
export type IAwareness = Awareness;
